
#include "HelloWorld_subscriber.h"
#include "HelloWorldSupport.h"
#include "HelloSvc.h"
#include <iostream>

extern "C" int subscriber_main(int domainId, int sample_count);
extern int subscriber_shutdown(DDSDomainParticipant *participant);

HelloSvc::HelloSvc(const tstring& tsName, const tstring& tsEventLogDLL) :
	WinSvc(tsName, tsEventLogDLL),
	m_nDomainID(0),
	m_nSampleCount(0),
	participant(0),
	subscriber(0),
	topic(0),
	reader_listener(0),
	reader(0)
{
}

HelloSvc::~HelloSvc()
{
}

void HelloSvc::OnRun()
{
	GetLogger() << "Server up and running on domain:" << m_nDomainID << std::endl;
	SubscriberMain(m_nDomainID, m_nSampleCount);
};

void HelloSvc::OnStop()
{
	GetLogger() << "Stopping Server..." << std::endl;
	
	/* Delete all entities */

	(void) SubscriberShutdown(participant);
	delete reader_listener;

	GetLogger() << "Stopping Server... Done" << std::endl;
};

int HelloSvc::SubscriberMain(int domainId, int sample_count)
{
	DDS_ReturnCode_t retcode;
	const char *type_name;
	int count = 0;
	struct DDS_Duration_t receive_period = {4,0};
	int status = 0;

	/* Removing shared memory support */
	DDS_DomainParticipantQos participant_qos;

	DDSTheParticipantFactory->get_default_participant_qos(participant_qos);

	participant_qos.transport_builtin.mask = DDS_TRANSPORTBUILTIN_UDPv4;

	participant_qos.discovery.initial_peers.length(2);
	participant_qos.discovery.initial_peers[0] =
			 DDS_String_dup("builtin.udpv4://127.0.0.1");
	participant_qos.discovery.initial_peers[1] =
			DDS_String_dup("builtin.udpv4://239.225.0.1");

	/* To customize participant QoS, use 
	DDSTheParticipantFactory->get_default_participant_qos() */
	participant = DDSTheParticipantFactory->create_participant(
		domainId, participant_qos /*DDS_PARTICIPANT_QOS_DEFAULT*/, 
		NULL /* listener */, DDS_STATUS_MASK_NONE);
	if (participant == NULL) {
		GetLogger() << "create_participant error" << std::endl;
		SubscriberShutdown(participant);
		return -1;
	}

	/* To customize subscriber QoS, use
	participant->get_default_subscriber_qos() */
	subscriber = participant->create_subscriber(
		DDS_SUBSCRIBER_QOS_DEFAULT, NULL /* listener */, DDS_STATUS_MASK_NONE);
	if (subscriber == NULL) {
		GetLogger() << "create_subscriber error" << std::endl;
		SubscriberShutdown(participant);
		return -1;
	}

	/* Register type before creating topic */
	type_name = HelloWorldTypeSupport::get_type_name();
	retcode = HelloWorldTypeSupport::register_type(
		participant, type_name);
	if (retcode != DDS_RETCODE_OK) {
		GetLogger() << "register_type error " 
					<< retcode 
					<< std::endl;
		SubscriberShutdown(participant);
		return -1;
	}

	/* To customize topic QoS, use
	participant->get_default_topic_qos() */
	topic = participant->create_topic(
		"Example HelloWorld",
		type_name, DDS_TOPIC_QOS_DEFAULT, NULL /* listener */,
		DDS_STATUS_MASK_NONE);
	if (topic == NULL) {
		GetLogger() << "create_topic error" << std::endl;
		SubscriberShutdown(participant);
		return -1;
	}

	/* Create data reader listener */
	reader_listener = new HelloWorldListener(this);
	if (reader_listener ==  NULL) {
		GetLogger() << "listener instantiation error" << std::endl;
		SubscriberShutdown(participant);
		return -1;
	}

	/* To customize data reader QoS, use
	subscriber->get_default_datareader_qos() */
	reader = subscriber->create_datareader(
		topic, DDS_DATAREADER_QOS_DEFAULT, reader_listener,
		DDS_STATUS_MASK_ALL);
	if (reader == NULL) {
		GetLogger() << "create_datareader error" << std::endl;
		SubscriberShutdown(participant);
		delete reader_listener;
		return -1;
	}

	return status;
}

int HelloSvc::SubscriberShutdown(DDSDomainParticipant *participant)
{
	DDS_ReturnCode_t retcode;
	int status = 0;

	if (participant != NULL) {
		retcode = participant->delete_contained_entities();
		if (retcode != DDS_RETCODE_OK) {
			printf("delete_contained_entities error %d\n", retcode);
			status = -1;
		}

		retcode = DDSTheParticipantFactory->delete_participant(participant);
		if (retcode != DDS_RETCODE_OK) {
			printf("delete_participant error %d\n", retcode);
			status = -1;
		}
	}

	/* RTI Data Distribution Service provides finalize_instance() method on
	domain participant factory and finalize() method on type support for
	people who want to release memory used by the participant factory and
	type support singletons. Uncomment the following block of code for
	clean destruction of the singletons. */
	/*
	HelloWorldTypeSupport::finalize();

	retcode = DDSDomainParticipantFactory::finalize_instance();
	if (retcode != DDS_RETCODE_OK) {
	printf("finalize_instance error %d\n", retcode);
	status = -1;
	}
	*/
	return status;	
}

void HelloSvc::SetDomainID(int domainID)
{
	m_nDomainID = domainID;
}

void HelloSvc::SetSampleCount(int sample_count)
{
	m_nSampleCount = sample_count;
}