
//#include "stdafx.h"
#include "WinService.h"
//#include "service_msg.h"
#include "windows.h"
#include <strsafe.h>	// for StringCchPrintf
#include <iostream>

// ----------------------------------------------------------------------------
// Implementation of WinSvc class
// ----------------------------------------------------------------------------

// Define static member variable
WinSvc* WinSvc::m_pThis = 0;

// ----------------------------------------------------------------------------
// WINAPI interface functions

#define MSG_WINDOWS_ERROR                ((DWORD)0xC0020001L)

// Function to handle control requests
void WINAPI WinSvc::SvcControlHandler(DWORD dwControl)
{
	try {
		switch (dwControl)
		{
		case SERVICE_CONTROL_STOP:
			m_pThis->SvcReportStatus( SERVICE_STOP_PENDING, NO_ERROR, 0 );
			m_pThis->OnStop();
			SetEvent(m_pThis->m_hSvcStopEvent);
			m_pThis->SvcReportStatus( SERVICE_STOP_PENDING, NO_ERROR, 0 );
			break;

		case SERVICE_CONTROL_PAUSE:
			m_pThis->SvcReportStatus( SERVICE_PAUSE_PENDING, NO_ERROR, 0 );
			m_pThis->OnPause();
			m_pThis->SvcReportStatus( SERVICE_PAUSED, NO_ERROR, 0 );
			break;

		case SERVICE_CONTROL_CONTINUE:
			m_pThis->SvcReportStatus( SERVICE_CONTINUE_PENDING, NO_ERROR, 0 );
			m_pThis->OnContinue();
			m_pThis->SvcReportStatus( SERVICE_RUNNING, NO_ERROR, 0 );
			break;

		case SERVICE_CONTROL_INTERROGATE:
			m_pThis->SvcReportStatus( SERVICE_RUNNING, NO_ERROR, 0 );
			m_pThis->OnInterrogate();
			m_pThis->SvcReportStatus( SERVICE_RUNNING, NO_ERROR, 0 );
			break;

		case SERVICE_CONTROL_SHUTDOWN:
			m_pThis->OnShutdown();
			break;

		case SERVICE_CONTROL_PARAMCHANGE:
			m_pThis->SvcReportStatus( SERVICE_RUNNING, NO_ERROR, 0 );
			m_pThis->OnParamChange();
			m_pThis->SvcReportStatus( SERVICE_RUNNING, NO_ERROR, 0 );
			break;

		case SERVICE_CONTROL_NETBINDADD:
			m_pThis->SvcReportStatus( SERVICE_RUNNING, NO_ERROR, 0 );
			m_pThis->OnNetBindAdd();
			m_pThis->SvcReportStatus( SERVICE_RUNNING, NO_ERROR, 0 );
			break;

		case SERVICE_CONTROL_NETBINDREMOVE:
			m_pThis->SvcReportStatus( SERVICE_RUNNING, NO_ERROR, 0 );
			m_pThis->OnNetBindRemove();
			m_pThis->SvcReportStatus( SERVICE_RUNNING, NO_ERROR, 0 );
			break;

		case SERVICE_CONTROL_NETBINDENABLE:
			m_pThis->SvcReportStatus( SERVICE_RUNNING, NO_ERROR, 0 );
			m_pThis->OnNetBindEnable();
			m_pThis->SvcReportStatus( SERVICE_RUNNING, NO_ERROR, 0 );
			break;

		case SERVICE_CONTROL_NETBINDDISABLE:
			m_pThis->SvcReportStatus( SERVICE_RUNNING, NO_ERROR, 0 );
			m_pThis->OnNetBindDisable();
			m_pThis->SvcReportStatus( SERVICE_RUNNING, NO_ERROR, 0 );
			break;

		case SERVICE_CONTROL_DEVICEEVENT:
			m_pThis->SvcReportStatus( SERVICE_RUNNING, NO_ERROR, 0 );
			m_pThis->OnDeviceEvent();
			m_pThis->SvcReportStatus( SERVICE_RUNNING, NO_ERROR, 0 );
			break;

		case SERVICE_CONTROL_HARDWAREPROFILECHANGE:
			m_pThis->SvcReportStatus( SERVICE_RUNNING, NO_ERROR, 0 );
			m_pThis->OnHardwareProfileChange();
			m_pThis->SvcReportStatus( SERVICE_RUNNING, NO_ERROR, 0 );
			break;

		case SERVICE_CONTROL_POWEREVENT:
			m_pThis->SvcReportStatus( SERVICE_RUNNING, NO_ERROR, 0 );
			m_pThis->OnPowerEvent();
			m_pThis->SvcReportStatus( SERVICE_RUNNING, NO_ERROR, 0 );
			break;

		case SERVICE_CONTROL_SESSIONCHANGE:
			m_pThis->SvcReportStatus( SERVICE_RUNNING, NO_ERROR, 0 );
			m_pThis->OnSessionChange();
			m_pThis->SvcReportStatus( SERVICE_RUNNING, NO_ERROR, 0 );
			break;

		default:
			m_pThis->GetLogger() << "Invalid service control request: " 
								 << dwControl 
								 << std::endl; 
			break;
		}
	}
	catch(ExWinAPI& ex)
	{
		m_pThis->GetLogger() << ex.GetErrMsg() << std::endl;
		m_pThis->LogEvent(m_pThis->m_tsName, 
						  MSG_WINDOWS_ERROR, 
						  EVENTLOG_ERROR_TYPE, 
						  1, 
						  ex.GetErrMsg());
	}
}

// Entry point for SCM
void WINAPI WinSvc::ServiceMain(DWORD dwArgc, LPTSTR* lpszArgv)
{
	try {
		// Start service
		m_pThis->m_hSvcStatus = 
			RegisterServiceCtrlHandler(
				m_pThis->m_tsName.c_str(), 
				SvcControlHandler);

		if (!m_pThis->m_hSvcStatus)
			throw ExWinAPI();

		// Set service attributes
		m_pThis->m_SvcStatus.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
		m_pThis->m_SvcStatus.dwServiceSpecificExitCode = 0;

		// Inform SCM about our status
		m_pThis->SvcReportStatus(SERVICE_START_PENDING, NO_ERROR, 3000);

		// Create stop condition, if we fail to create the stop event,
		// we should never start in the first place!
		m_pThis->m_hSvcStopEvent = CreateEvent(0, TRUE, FALSE, 0);

		if (!m_pThis->m_hSvcStopEvent)
		{
			m_pThis->SvcReportStatus(SERVICE_STOPPED, NO_ERROR, 0);
			m_pThis->GetLogger() << "Failed to create stop condition, " 
								 << "not started" 
								 << std::endl;
			return;
		}

		// Inform SCM about our status (again!)
		m_pThis->SvcReportStatus(SERVICE_RUNNING, NO_ERROR, 0);

		// Call custom task
		m_pThis->OnRun();

		// Sleep, till we're terminated
		WaitForSingleObject(m_pThis->m_hSvcStopEvent, INFINITE);
		m_pThis->SvcReportStatus(SERVICE_STOPPED, NO_ERROR, 0);
	}
	catch(ExWinAPI& ex)
	{
		m_pThis->GetLogger() << ex.GetErrMsg() << std::endl;

		m_pThis->LogEvent(m_pThis->m_tsName, 
			MSG_WINDOWS_ERROR, 
			EVENTLOG_ERROR_TYPE, 
			1, 
			ex.GetErrMsg());
	}
}

// Handler called when we're running in a console window and CTRL-C is hit.
int WINAPI WinSvc::ExitRoutine(unsigned long ulControlType)
{
	if (ulControlType == CTRL_C_EVENT     || 
		ulControlType == CTRL_BREAK_EVENT || 
		ulControlType == CTRL_CLOSE_EVENT || 
		ulControlType == CTRL_SHUTDOWN_EVENT )
	{
		if (m_pThis->m_hExitEvent)
		{
			m_pThis->OnExit();
			SetEvent(m_pThis->m_hExitEvent);
		}
		return 1;
	}
	return 0;
}

// ----------------------------------------------------------------------------
// Construction and destruction

// Constructor; allocate local structures
WinSvc::WinSvc(const tstring& tsServiceName, 
			   const tstring& tsEventLogDLL,
			   const tstring& tsLogPath) :
	m_bSvcRunning(false),
	m_hSvcStopEvent(0),
	m_hExitEvent(0),
	m_hSvc(0),
	m_hSCM(0),
	m_hSvcStatus(0),
	m_pLogStream(0)
{
	// Make sure we can log all our errors
	OpenLogFile(tsLogPath);

	// Store this pointer to access WinSvc instance from Windows API functions
	m_pThis = this;

	// Set service name;
	m_tsName = tsServiceName;

	// Allocate memory for servicename to pass it to dispatch table
	TCHAR* pszName = new TCHAR[ (m_tsName.length() + 1) * sizeof(TCHAR) ];
	memset(pszName, 0, (m_tsName.length() + 1) * sizeof(TCHAR));
	memcpy(pszName, m_tsName.c_str(), m_tsName.length() * sizeof(TCHAR));

	// Create empty dispatch table
	m_pDispatchTable = new SERVICE_TABLE_ENTRY[2];
	memset(m_pDispatchTable, 0, sizeof(*m_pDispatchTable));

	// Fill dispatch table
	m_pDispatchTable[0].lpServiceName = pszName;
	m_pDispatchTable[0].lpServiceProc = ServiceMain;

	// Register event source
	AddEventSourceToRegistry(TEXT("Application"), 
							 m_tsName, 
							 tsEventLogDLL, 
							 100);
}

// Destructor; cleanup local structures and close handles
WinSvc::~WinSvc()
{
	delete [] m_pDispatchTable[0].lpServiceName;
	delete [] m_pDispatchTable;

	if (m_hSvcStopEvent)
		CloseHandle(m_hSvcStopEvent);

	if (m_hExitEvent)
		CloseHandle(m_hExitEvent);

	if (m_hSvc)
		CloseServiceHandle(m_hSvc);

	if (m_hSCM)
		CloseServiceHandle(m_hSCM);

	if (m_pLogStream)
	{
		m_pLogStream->close();
		delete m_pLogStream;
	}
}

// ----------------------------------------------------------------------------
// Member functions

// Connect main thread service process to the SCM. This causes the thread to 
// be the service control dispatcher thread for the calling process.
void WinSvc::SvcStartDispatcher()
{ 
	if (!StartServiceCtrlDispatcher(m_pDispatchTable))
		throw ExWinAPI();
};

// Updates the SCM's status information for the calling service.
void WinSvc::SvcReportStatus(DWORD dwState, DWORD dwExitCode, DWORD dwHint)
{
	static DWORD dwCheckPoint = 1;

	m_SvcStatus.dwCurrentState  = dwState;
	m_SvcStatus.dwWin32ExitCode = dwExitCode;
	m_SvcStatus.dwWaitHint      = dwHint;

	if (SERVICE_START_PENDING == m_SvcStatus.dwCurrentState)
	{
		m_SvcStatus.dwControlsAccepted = 0;
	}
	else
	{
		m_SvcStatus.dwControlsAccepted = SERVICE_ACCEPT_STOP;
	}

	if ((SERVICE_RUNNING == m_SvcStatus.dwCurrentState) ||
		(SERVICE_STOPPED == m_SvcStatus.dwCurrentState))
	{
		m_SvcStatus.dwCheckPoint = 0;
	}
	else
	{
		m_SvcStatus.dwCheckPoint++;
	}
	
	if (!SetServiceStatus(m_hSvcStatus, &m_SvcStatus))
		throw ExWinAPI();
}

// Replace default CTR_C handler by our own handler
bool WinSvc::InstallExitHandler()
{
	m_hExitEvent = CreateEvent(0, FALSE, FALSE, 0);

	if (!m_hExitEvent)
		return false;

	if (!SetConsoleCtrlHandler(ExitRoutine, true))
	{
		GetLogger() << "Failed to install console CTRL-C handler" << std::endl;
		CloseHandle(m_hExitEvent);
		return false;
	}
	return true;
}

// Set service description of service. 
// This setting is displayed by the SCM.
void WinSvc::SetDescription(const tstring& tsDescription)
{
	m_tsDescription = tsDescription;

	SERVICE_DESCRIPTION sd;

	// Change the service description
	TCHAR* pszDesc = new TCHAR[ (m_tsDescription.length() + 1) * sizeof(TCHAR) ];
	memset(pszDesc, 0, (m_tsDescription.length() + 1) * sizeof(TCHAR));
	memcpy(pszDesc, m_tsDescription.c_str(), m_tsDescription.length() * sizeof(TCHAR));

	sd.lpDescription = pszDesc;

	if (!ChangeServiceConfig2(
			GetSvcHandle(),			// handle to service
			SERVICE_CONFIG_DESCRIPTION,		// change: description
			&sd)) 							// new description
	{
		delete [] pszDesc;
		throw ExWinAPI();
	}
	delete [] pszDesc;
}

// Establishes a connection to the service control manager on the specified 
// computer and opens the specified service control manager database.
const SC_HANDLE& WinSvc::GetSCManagerHandle()
{
	if (!m_hSCM)
	{
		m_hSCM = OpenSCManager(
					NULL,
					SERVICES_ACTIVE_DATABASE,
					SC_MANAGER_CREATE_SERVICE);

		if (!m_hSCM)
		{
			ExWinAPI ex;
			GetLogger() << ex.GetErrMsg() << std::endl;
			LogEvent(m_pThis->m_tsName, 
					 MSG_WINDOWS_ERROR, 
					 EVENTLOG_ERROR_TYPE, 
					 1, 
					 ex.GetErrMsg());
		}
	}
	return m_hSCM;
}

// Open the (existing) service.
const SC_HANDLE& WinSvc::GetSvcHandle()
{
	if (!m_hSvc)
	{
		m_hSvc = OpenService( 
					GetSCManagerHandle(), 
					m_tsName.c_str(),
					SERVICE_ALL_ACCESS);
		if (!m_hSvc)
		{
			ExWinAPI ex;
			GetLogger() << ex.GetErrMsg() << std::endl;

			LogEvent(m_pThis->m_tsName, 
				MSG_WINDOWS_ERROR, 
				EVENTLOG_ERROR_TYPE, 
				1, 
				ex.GetErrMsg());
		}
	}
	return m_hSvc;
}

// Set failure actions flag setting of service. 
// This setting determines when failure actions are to be executed.
void WinSvc::SetFailActions(DWORD dwResetPeriod, int nTimeOut)
{
	// Configure service with ServiceTimeout
	const int nNumActions(1);

	SERVICE_FAILURE_ACTIONS sfa;
	SC_ACTION arrayActions[nNumActions];

	// Time to wait before performing specified action (milliseconds)
	arrayActions[0].Delay = nTimeOut * 1000 * 60; 
	arrayActions[0].Type  = SC_ACTION_RESTART;

	// Change the failure actions
	sfa.dwResetPeriod = dwResetPeriod;
	sfa.lpRebootMsg   = 0;
	sfa.lpCommand     = 0;
	sfa.cActions      =	nNumActions; // Number of elements in the lpsaActions array
	sfa.lpsaActions   = arrayActions;

	if (!ChangeServiceConfig2(
		GetSvcHandle(),			 // handle to service
		SERVICE_CONFIG_FAILURE_ACTIONS,  // change: failure actions
		&sfa))
	{
		throw ExWinAPI();
	}
}

// ----------------------------------------------------------------------------
// Creation and removal (add/remove service to/from registry)

// Creates a service object and adds it to the specified service control manager database.
bool WinSvc::Create(
				const tstring& tsDescription,
				const tstring& tsFullExePath,
				int nTimeOut,
				DWORD dwResetPeriod)
{
	bool bRetVal = true;

	try {
		SC_HANDLE m_SvcHnd = 
			CreateService( 
				GetSCManagerHandle(), 
				m_tsName.c_str(), 
				m_tsName.c_str(),
				SERVICE_ALL_ACCESS,
				SERVICE_WIN32_OWN_PROCESS,
				SERVICE_AUTO_START,
				SERVICE_ERROR_IGNORE,
				tsFullExePath.c_str(),
				NULL, 
				NULL, 
				NULL, 
				NULL, 
				NULL); 

		if (!m_SvcHnd)
		{
			throw ExWinAPI();
		}
		else
		{
			SetDescription(tsDescription);
			SetFailActions(dwResetPeriod, nTimeOut);
		}
	} catch(ExWinAPI& ex) {
		GetLogger() << ex.GetErrMsg() << std::endl; 	

		LogEvent(m_pThis->m_tsName, 
			MSG_WINDOWS_ERROR, 
			EVENTLOG_ERROR_TYPE, 
			1, 
			ex.GetErrMsg());
		bRetVal = false;
	}
	return bRetVal;
}

// Marks the service for deletion from the service control manager database.
bool WinSvc::Remove()
{
	bool bRetVal = true;

	try {
		if (!DeleteService( GetSvcHandle() ))
			throw ExWinAPI();
	}
	catch(ExWinAPI& ex)
	{
		GetLogger() << ex.GetErrCode() << TEXT(": ") << ex.GetErrMsg() << std::endl;

		LogEvent(m_pThis->m_tsName, 
			MSG_WINDOWS_ERROR, 
			EVENTLOG_ERROR_TYPE, 
			1, 
			ex.GetErrMsg());
		bRetVal = false;
	}
	return bRetVal;
};

// Check if the service is running (as service)
bool WinSvc::IsSvcRunning()
{
	// Set initial state to stopped in case service was not installed
	m_SvcStatus.dwCurrentState = SERVICE_STOPPED;

	// Get status information of service
	(void) ControlService(GetSvcHandle(), SERVICE_CONTROL_INTERROGATE, &m_SvcStatus);
	
	m_bSvcRunning = (SERVICE_STOPPED != m_SvcStatus.dwCurrentState);
	return m_bSvcRunning;
}

// Start service by calling SCM Dispatcher, if mode is equal to WINSVC_CONSOLE
// and the service is not running an exit handler is installed and all output
// is redirected to a console window.
bool WinSvc::Start(etWinSvcMode etSvcMode)
{
	m_etSvcMode = etSvcMode;
	bool bRetVal = true;
	try {
		if (m_etSvcMode == WINSVC_SERVICE)
		{
			SvcStartDispatcher();
		}
		else
		{
			if (!IsSvcRunning())
			{
				if (InstallExitHandler())
				{
					std::cout << "Entering debug mode, press CTRL-C to exit" << std::endl << std::endl;
					OnRun();
					std::cout << "Awaiting data..." << std::endl << std::endl;
					WaitForSingleObject(m_hExitEvent, INFINITE);
					OnStop();
					std::cout << "Debug mode aborted!" << std::endl;
				}
			}
			else
				GetLogger() << "Aborted, already running as service" << std::endl;
		}
	} catch (ExWinAPI& ex) {
		GetLogger() << ex.GetErrMsg() << std::endl;

		LogEvent(m_pThis->m_tsName, 
			MSG_WINDOWS_ERROR, 
			EVENTLOG_ERROR_TYPE, 
			1, 
			ex.GetErrMsg());
		bRetVal = false;
	}
	return bRetVal;
}
// ----------------------------------------------------------------------------
// Format message for last errorcode inside exception
tstring ExWinAPI::GetErrMsg()
{
	LPTSTR lpMsgBuf;

    FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER | 
        FORMAT_MESSAGE_FROM_SYSTEM |
        FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,
        m_ErrCode,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        (LPTSTR) &lpMsgBuf,
        0, NULL );

	tstring tsErrMsg = tstring(lpMsgBuf);
	LocalFree(lpMsgBuf);
	return tsErrMsg;
};

// ----------------------------------------------------------------------------

// Note: This function can only be used when AddEventSourceToRegistry is called
// to register the event source. This need to be once per machine, because the
// information is stored as non volatile in the registry.
void WinSvc::LogEvent(
			  const tstring& tsEventSourceName,
			  DWORD dwEventID,
			  WORD  wEventLogType,
			  WORD  wNumParams,
			  const tstring& tsParamData)
{
	// TODO: Support multiple parameters (use vector<tstring*> for example)
	// Set number of parameters to 1, no matter what is passed since we only
	// support one at the moment!
	wNumParams = 1;

	TCHAR* pszBuffer = 0;
	HANDLE hEventSource = 0;

	try {
		hEventSource = 
					RegisterEventSource(
						NULL,						// Use local computer
						tsEventSourceName.c_str());	// Event sourcename

		if (!hEventSource)
			throw ExWinAPI();

		pszBuffer = new TCHAR[ (tsParamData.length() + 1) * sizeof(TCHAR) ];
		memset(pszBuffer, 0, (tsParamData.length() + 1) * sizeof(TCHAR));
		memcpy(pszBuffer, tsParamData.c_str(), tsParamData.length() * sizeof(TCHAR));

		LPCTSTR pszParamData = pszBuffer;

		// Report the event
		if (!ReportEvent(
				hEventSource,			// Event log handle
				wEventLogType,			// Event type,
				0,						// Event category
				dwEventID,				// Event identifier
				NULL,					// No user security identifier
				wNumParams,				// Number of substitution strings
				0,						// No data
				&pszParamData,			// Pointer to strings
				NULL))					// No data
		{
			throw ExWinAPI();
		}
		
		if (!DeregisterEventSource(hEventSource))
			throw ExWinAPI();

		delete [] pszBuffer;

	} catch(ExWinAPI& ex) {
		
		GetLogger() << ex.GetErrMsg() << std::endl;

		if (pszBuffer)
			delete [] pszBuffer;

		// Try to deregister eventsource one more time
		if (hEventSource)
			DeregisterEventSource(hEventSource);
	}
}

// Add event source to registry to able to use EventLogger
// Returns without doing anything when key already exists in registry.
void WinSvc::AddEventSourceToRegistry(
				const tstring& tsLogName,
				const tstring& tsSourceName,
				const tstring& tsFullDllName,
				DWORD dwCategoryNum)
{
	// If nothing to add, return immediately
	if (tsFullDllName.empty())
		return;

	// ------------------------------------------------------------------------
	// Create the event source as a subkey of the log. 
	// ------------------------------------------------------------------------
	tostringstream tosSrcKey;
	tosSrcKey	<< TEXT("SYSTEM\\CurrentControlSet\\Services\\EventLog\\")
				<< tsLogName 
				<< "\\" 
				<< tsSourceName;
	
	DWORD dwDisposition;
	HKEY hkKey = 0;
	BYTE* lpData = 0;

	LONG lErrorCode = 
			RegCreateKeyEx(
				HKEY_LOCAL_MACHINE,			// Handle to open registry key
				tosSrcKey.str().c_str(),	// Name of subkey to create
				0,							// Reserved
				NULL,						// User defined classtype of this key
				REG_OPTION_NON_VOLATILE,	// Value preserved upon system restart
				KEY_WRITE,					// Write access rights
				NULL,						// Default security descriptor
				&hkKey,						// Handle to opened key
				&dwDisposition);			// Extra info about creation

	if (lErrorCode == ERROR_SUCCESS)
	{
		if (dwDisposition == REG_OPENED_EXISTING_KEY)
		{
			(void) RegCloseKey(hkKey);
			// Key was already created, we're done
			return;
		}
	}
	else
	{
		GetLogger() << "Could not create the registry key: " 
			<< tosSrcKey.str().c_str()
			<< ", error code: "
			<< lErrorCode
			<< std::endl;
	}

	// ------------------------------------------------------------------------
	// Set the name of the message file. 
	// ------------------------------------------------------------------------

	// Allocate memory for data
	lpData = new BYTE[ (tsFullDllName.length() + 1) * sizeof(TCHAR) ];
	memset(lpData, 0, (tsFullDllName.length() + 1) * sizeof(TCHAR));
	memcpy(	lpData, 
			tsFullDllName.c_str(), 
			tsFullDllName.length() * sizeof(TCHAR));

	DWORD dwDataSize = 
			static_cast<DWORD>(tsFullDllName.length()+1) * sizeof(TCHAR);

	lErrorCode = 
		RegSetValueEx(
			hkKey,						// Handle to open registry key
			TEXT("EventMessageFile"),	// Name of value to be set 
			0,							// Reserved
			REG_EXPAND_SZ,				// Value is null-terminated string 
			lpData,						// Data to be stored
			dwDataSize);				// Size of data

	if (lErrorCode != ERROR_SUCCESS)
	{
		GetLogger() << "Could not set the event message file: "
					<< " to "
					<< tsFullDllName.c_str()
					<< std::endl;

		RegCloseKey(hkKey); 
		delete [] lpData;
		lpData = 0;
		return;
	}

	// ------------------------------------------------------------------------
	// Set the supported event types. 
	// ------------------------------------------------------------------------
	DWORD dwData = 
		EVENTLOG_ERROR_TYPE | EVENTLOG_WARNING_TYPE | EVENTLOG_INFORMATION_TYPE; 

	lErrorCode = 
		RegSetValueEx(
		hkKey,								// Handle to open registry key
		TEXT("TypesSupported"),				// Name of value to be set 
		0,									// Reserved
		REG_DWORD,							// Value is null-terminated string 
		reinterpret_cast<LPBYTE>(&dwData),	// Data to be stored
		sizeof(DWORD));						// Size of data
	
	if (lErrorCode != ERROR_SUCCESS)
	{
		printf("Could not set the supported types."); 
		RegCloseKey(hkKey); 
		delete [] lpData;
		lpData = 0;
		return;
	}


	// ------------------------------------------------------------------------
	// Set the category message file
	// ------------------------------------------------------------------------

	dwDataSize = static_cast<DWORD>(tsFullDllName.length()+1) * sizeof(TCHAR);

	lErrorCode = 
		RegSetValueEx(
			hkKey,							// Handle to open registry key
			TEXT("CategoryMessageFile"),	// Name of value to be set 
			0,								// Reserved
			REG_EXPAND_SZ,					// Value is null-terminated string 
			lpData,							// Data to be stored
			dwDataSize);					// Size of data

	if (lErrorCode != ERROR_SUCCESS)
	{
		printf("Could not set the category message file."); 
		RegCloseKey(hkKey); 
		delete [] lpData;
		lpData = 0;
		return;
	}

	// ------------------------------------------------------------------------
	// Set the number of categories.
	// ------------------------------------------------------------------------
	lErrorCode = 
		RegSetValueEx(
			hkKey,								// Handle to open registry key
			TEXT("CategoryCount"),				// Name of value to be set 
			0,									// Reserved
			REG_DWORD,							// Value is null-terminated string 
												// Data to be stored
			reinterpret_cast<LPBYTE>(&dwCategoryNum),	
			sizeof(DWORD));						// Size of data

	if (lErrorCode != ERROR_SUCCESS)
	{
		printf("Could not set the category count."); 
		RegCloseKey(hkKey); 
		delete [] lpData;
		lpData = 0;
		return;
	}
	RegCloseKey(hkKey); 
	delete [] lpData;
	return;
}

// ----------------------------------------------------------------------------

// TODO: log to eventlog
// LogEvent(m_tsName, 0x01, EVENTLOG_WARNING_TYPE, 0, NULL);

// ----------------------------------------------------------------------------
// Retrieve a reference to the log-stream
tostream& WinSvc::GetLogger()
{
	if (m_pLogStream)
		return (*m_pLogStream);
	else
	{
#ifdef UNICODE
		return std::wcout;
#else
		return std::cout;
#endif
	}
};

// ----------------------------------------------------------------------------
// Open logfile
void WinSvc::OpenLogFile(const tstring& tsLogPath)
{
	if (!tsLogPath.empty())
	{
		tstring tsFileName = tsLogPath + m_tsName + TEXT(".log");

		if (m_pLogStream)
		{
			m_pLogStream->close();
			delete m_pLogStream;
			m_pLogStream = 0;
		}
		m_pLogStream = new tofstream(tsFileName.c_str(), std::ios_base::app);
	}
}

// ----------------------------------------------------------------------------
